/*
 * Serial Flash Support 
 */
#include <common.h>
#include <command.h>

#if (CONFIG_COMMANDS & CFG_CMD_SFLASH)

#include <serial_flash.h>

extern serial_flash_info_t sflash_info;

static int s_abbrev_spec (char *str, serial_flash_info_t *pinfo, int *psf, int *psl)
{
	serial_flash_info_t *fp;
	ulong bank, first, last;
	char *p, *ep;

	if ((p = strchr(str, ':')) == NULL)
		return 0;
	*p++ = '\0';

	/* NOTE TO SELF: BANKS ARE NOT IMPLEMENTED O_o */
	bank = simple_strtoul(str, &ep, 10);
	if (ep == str || *ep != '\0' || bank != 1 || 
	   ((fp = pinfo)->serial_flash_id == SERIAL_FLASH_UNKNOWN))
		return -1;

	str = p;
	if ((p = strchr(str, '-')) != NULL)
		*p++ = '\0';
	
	first = simple_strtoul(str, &ep, 10);
	if (ep == str || *ep != '\0' || first >= fp->sector_count)
		return -1;
	
	if (p != NULL) {
		last = simple_strtoul(p, &ep, 10);
		if (ep == p || *ep != '\0' || last < first
			    || last >= fp->sector_count)
			return -1;
	}
	else {
		last = first;
	}

	*psf = first;
	*psl = last;
	return 1;	
}


int do_sflinfo (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	/* NOTE: I DIDNT IMPLEMENT SERIAL FLASH BANKS YET */
	serial_flash_print_info();
	return 0;
}

int do_sflerase (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	ulong addr_first, addr_last;
	int n, sect_first, sect_last;
	int rcode = 0;

	if (argc < 2) {
		printf("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	if (strcmp(argv[1], "all") == 0) {
		puts ("Erase entire serial flash ");
		rcode = serial_flash_bulk_erase();
		serial_flash_perror(rcode);
		return rcode;
	}

	if ((n = s_abbrev_spec(argv[1], &sflash_info, &sect_first, &sect_last)) != 0) {
		if(n < 0) {
			puts ("Bad sector specification\n");
			return 1;
		}
		printf ("Erase serial flash sector-range: %d-%d ", sect_first, sect_last);
		rcode = serial_flash_erase(sect_first, sect_last);
		serial_flash_perror(rcode);
		return rcode;
	}

	if (argc != 3) {
		printf("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	addr_first = simple_strtoul (argv[1], NULL, 16);
	addr_last  = simple_strtoul (argv[2], NULL, 16);

	if (addr_first >= addr_last) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
	
	puts ("Erase serial flash ");
	rcode = serial_flash_sect_erase(addr_first, addr_last);
	serial_flash_perror(rcode);
	return rcode;
}


int do_srealprotect (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	unsigned int bp;
	int rcode = 0;

	if (argc != 2) {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}
	

	/* HOW DO I STOP THE USER FROM ENTERING RUBBISH?!
	   THE STUPID FUNCTION CONVERTS EVERYTHING! */	
	bp = (unsigned int) simple_strtoul(argv[1], NULL, 10);

	rcode = serial_flash_block_protect((uchar) bp);
	return rcode;
}


int do_sprotect (cmd_tbl_t *cmdtp, int flag, int argc, char *argv[])
{
	int i, p, rcode;
	int n, sect_first, sect_last;
	ulong addr_first, addr_last;

	if(argc < 3) {
		printf("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	if (strcmp(argv[1], "off") == 0) {
		p = 0;
	} 
	else if (strcmp(argv[1], "on") == 0) {
		p = 1;
	} else {
		printf ("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	if(strcmp(argv[2], "all") == 0) {
		printf("%sProtect entire serial flash\n",
			p ? "" : "Un-");
		for (i = 0; i < sflash_info.sector_count; i++)
			sflash_info.sector_protect[i] = p ? PROTECTED : NOT_PROTECTED;
		return 0;
	}

	if((n = s_abbrev_spec(argv[2], &sflash_info, &sect_first, &sect_last)) != 0) {
		if(n < 0) {
			puts("Bad sector specification\n");
			return 1;
		}
		printf("%sProtect serial flash sectors %d to %d\n", 
			p ? "" : "Un-",
			sect_first, sect_last);
		for (i = sect_first; i <= sect_last; i++)
			sflash_info.sector_protect[i] = p ? PROTECTED : NOT_PROTECTED;
		return 0;
	}

	if(argc != 4) {
		printf("Usage:\n%s\n", cmdtp->usage);
		return 1;
	}

	addr_first = simple_strtoul(argv[2], NULL, 16);
	addr_last  = simple_strtoul(argv[3], NULL, 16);

	rcode = serial_flash_protect(p, addr_first, addr_last);
	if(rcode != 0) {
		puts("Bad address specification\n");
		return 1;
	}	
	printf("%sProtect serial flash sectors\n", p ? "" : "Un-");
	return 0;
}

/**************************************************/

U_BOOT_CMD(
	sflinfo,      1,    1,    do_sflinfo,
	"sflinfo  - print SERIAL-FLASH memory information\n",
	"\n    - print information for the entire serial FLASH memory\n"
);

U_BOOT_CMD(
	serase,       3,    1,   do_sflerase,
	"serase   - erase SERIAL-FLASH memory\n",
	"start end\n"
	"    - erase SERIAL-FLASH from addr 'start' to addr 'end'\n"
	"serase N:SF[-SL]\n    - erase sectors SF-SL in SERIAL-FLASH bank #N\n"
	"serase all\n    - erase all SERIAL-FLASH memory\n"
);

U_BOOT_CMD(
	srealprotect,  2,   1,   do_srealprotect,
	"srealprotect - adjust SERIAL-FLASH write block protection\n",
	"block\n"
	"    - protect block of sectors specified by 'block'\n"
);

U_BOOT_CMD(
	sprotect,      4,   1,   do_sprotect,
	"sprotect - enable or disable SERIAL-FLASH write protection\n",
	"on  start end\n"
	"    - protect SERIAL FLASH from addr 'start' to addr 'end'\n"
	"protect on N:SF[-SL]\n"
	"    - protect sectors SF-SL in SERIAL FLASH\n"
	"protect on all\n    - protect all SERIAL FLASH\n"
	"off start end\n"
	"    - make SERIAL FLASH writable from addr 'start' to addr 'end'\n"
	"protect off N:SF[-SL]\n"
	"    - make sectors SF-SL in SERIAL FLASH writable\n"
	"protect off all\n    - make all SERIAL FLASH writable\n"
);
#endif /* CFG_CMD_SERIAL_FLASH */ 

