/*
 * INCA-IP internal switch ethernet driver.
 *
 * (C) Copyright 2003-2004
 * Wolfgang Denk, DENX Software Engineering, wd@denx.de.
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston,
 * MA 02111-1307 USA
 */


#include <common.h>

#if (CONFIG_COMMANDS & CFG_CMD_NET) && defined(CONFIG_NET_MULTI) \
	&& defined(CONFIG_INCA_IP_SWITCH)
#include <config.h>
#include <malloc.h>
#include <net.h>
#include <asm/inca-ip.h>
#include <asm/addrspace.h>


#define NUM_RX_DESC	PKTBUFSRX
#define NUM_TX_DESC	3
#define TOUT_LOOP	1000000


#define DELAY	udelay(10000)
  /* Sometimes the store word instruction hangs while writing to one
   * of the Switch registers. Moving the instruction into a separate
   * function somehow makes the problem go away.
   */
static void SWORD(volatile u32 * reg, u32 value)
{
	*reg = value;
}

#define DMA_WRITE_REG(reg, value) *((volatile u32 *)reg) = (u32)value;
#define DMA_READ_REG(reg, value)    value = (u32)*((volatile u32*)reg)
#define SW_WRITE_REG(reg, value)   \
	SWORD(reg, value);\
	DELAY;\
	SWORD(reg, value);

#define SW_READ_REG(reg, value)	   \
	value = (u32)*((volatile u32*)reg);\
	DELAY;\
	value = (u32)*((volatile u32*)reg);

#define INCA_DMA_TX_POLLING_TIME	0x07
#define INCA_DMA_RX_POLLING_TIME	0x07

#define INCA_DMA_TX_HOLD		0x80000000
#define INCA_DMA_TX_EOP			0x40000000
#define INCA_DMA_TX_SOP			0x20000000
#define INCA_DMA_TX_ICPT		0x10000000
#define INCA_DMA_TX_IEOP		0x08000000

#define INCA_DMA_RX_C			0x80000000
#define INCA_DMA_RX_SOP			0x40000000
#define INCA_DMA_RX_EOP			0x20000000

#define INCA_SWITCH_PHY_SPEED_10H	0x1
#define INCA_SWITCH_PHY_SPEED_10F	0x5
#define INCA_SWITCH_PHY_SPEED_100H	0x2
#define INCA_SWITCH_PHY_SPEED_100F	0x6

/************************ Auto MDIX settings ************************/
#define INCA_IP_AUTO_MDIX_LAN_PORTS_DIR		INCA_IP_Ports_P1_DIR
#define INCA_IP_AUTO_MDIX_LAN_PORTS_ALTSEL	INCA_IP_Ports_P1_ALTSEL
#define INCA_IP_AUTO_MDIX_LAN_PORTS_OUT		INCA_IP_Ports_P1_OUT
#define INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX	16

#define WAIT_SIGNAL_RETRIES			100

#ifdef INCAIP_AUTO_MDIX
#define WAIT_LINK_RETRIES			5
#else
#define WAIT_LINK_RETRIES			2
#endif

#define LINK_RETRY_DELAY			3000  /* ms */
/********************************************************************/


/*************** MAC table access settings and structures *******************/
/* max. number of INCA-IP MAC entries within the switch MAC table */
#define INCA_IP_MAX_MAC_TABLE_ENTRIES 256


/* 
   CPU Access Control Register: write to MA table
   --> the MAC table index (hash value) has to be set by the user 
   # RA  = 1   request ack
   # RW  = 1   request write
   # AVA = 1   Access to MA table (0 --> access to VLAN)
*/
#define INCA_IP_Switch_CPU_ACTL_WR_MA_TABLE  ( INCA_IP_Switch_CPU_ACTL_RA | \
                                               INCA_IP_Switch_CPU_ACTL_RW | \
                                               INCA_IP_Switch_CPU_ACTL_AVA )
/* 
   CPU Access Control Register: read from MA table
   --> the MAC table index (hash value) has to be set by the user 
   # RA  = 1   request ack
   # AVA = 1   Access to MA table (0 --> access to VLAN)
   - RW  = 0   request read
*/
#define INCA_IP_Switch_CPU_ACTL_RD_MA_TABLE  ( INCA_IP_Switch_CPU_ACTL_RA | \
                                               INCA_IP_Switch_CPU_ACTL_AVA )
                                               
                                               

/* identify the available ports: PC, LAN, CPU */
#define INCA_IP_SWITCH_PORT_SELECT_PC      0
#define INCA_IP_SWITCH_PORT_SELECT_LAN     1
#define INCA_IP_SWITCH_PORT_SELECT_CPU     2                                               
                                               
/* for MAC table: decide between voice and management */
#define INCA_IP_SWITCH_PORT_SELECT_CPU_V   2
#define INCA_IP_SWITCH_PORT_SELECT_CPU_M   3                                               

/* bit defs for access a mac table entry */
#define INCA_IP_MAC_TABLE_ENTRY_PORT_PC_MASK       0x01
#define INCA_IP_MAC_TABLE_ENTRY_PORT_LAN_MASK      0x02
#define INCA_IP_MAC_TABLE_ENTRY_PORT_CPU_V_MASK    0x04
#define INCA_IP_MAC_TABLE_ENTRY_PORT_CPU_M_MASK    0x08
#define INCA_IP_MAC_TABLE_ENTRY_ALL_PORT_MASK      0x0F

#define INCA_IP_MAC_TABLE_ENTRY_CRITICAL_MASK      0x10
#define INCA_IP_MAC_TABLE_ENTRY_NO_CRITICAL_MASK   0x00

/* MAC table entry - control byte: 111x xxxx --> 0xE0 */
#define INCA_IP_MAC_TABLE_ENTRY_MA_ST_AGED_OUT_MASK      0xE0
#define INCA_IP_MAC_TABLE_ENTRY_MA_ST_STATIC_MASK        0x00
#define INCA_IP_MAC_TABLE_ENTRY_MA_ST_NORMAL_AGING_MASK  0x40



/* 
   union definition for the Switch CPU Access Data (CAD) register 
   --> access to the MAC table via CPU_ACTL register
*/
typedef union sw21553_cad_register
{
   u32   reg32 ;
   unsigned char    reg8[8] ;
} sw21553_CdaRegister ;
                                               
/* definition of an MAC entry used within the INCA-IP switch module */
typedef struct sw21553_mac_table_entry
{
   unsigned char MacAddr[6] ;
   unsigned char Cntrl ;
} sw21553_MacTableEntry ;
/********************************************************************/


typedef struct
{
	union {
		struct {
			volatile u32 HOLD		:1;
			volatile u32 ICpt		:1;
			volatile u32 IEop		:1;
			volatile u32 offset		:3;
			volatile u32 reserved0		:4;
			volatile u32 NFB		:22;
		}field;

		volatile u32 word;
	}params;

	volatile u32 nextRxDescPtr;

	volatile u32 RxDataPtr;

	union {
		struct {
			volatile u32 C			:1;
			volatile u32 Sop		:1;
			volatile u32 Eop		:1;
			volatile u32 reserved3		:12;
			volatile u32 NBT		:17;
		}field;

		volatile u32 word;
	}status;

} inca_rx_descriptor_t;


typedef struct
{
	union {
		struct {
			volatile u32 HOLD		:1;
			volatile u32 Eop		:1;
			volatile u32 Sop		:1;
			volatile u32 ICpt		:1;
			volatile u32 IEop		:1;
			volatile u32 reserved0		:5;
			volatile u32 NBA		:22;
		}field;

		volatile u32 word;
	}params;

	volatile u32 nextTxDescPtr;

	volatile u32 TxDataPtr;

	volatile u32 C			:1;
	volatile u32 reserved3		:31;

} inca_tx_descriptor_t;


static inca_rx_descriptor_t rx_ring[NUM_RX_DESC] __attribute__ ((aligned(16)));
static inca_tx_descriptor_t tx_ring[NUM_TX_DESC] __attribute__ ((aligned(16)));

static int tx_new, rx_new, tx_hold, rx_hold;
static int tx_old_hold = -1;
static int initialized	= 0;


static int inca_switch_init(struct eth_device *dev, bd_t * bis);
static int inca_switch_send(struct eth_device *dev, volatile void *packet, int length);
static int inca_switch_recv(struct eth_device *dev);
static void inca_switch_halt(struct eth_device *dev);
static void inca_init_switch_chip(void);
static void inca_dma_init(void);
static int inca_amdix(void);
static int inca_check_connections(void);

/*
   structure used for provide hash codes to modulate a MAC address
*/
typedef struct inca_ip_hash_codes
{
   u16   Code_0[3] ;
   u16   Code_1[3] ;
   u16   Code_2[3] ;
} IncaIpHashCodes ;
/* 
   codes for modulate the MAC address, used for calculate the MAC table index 
*/
IncaIpHashCodes   IncaIpMacModulateCodes = 
{ 
   {0xAAAA, 0xCCCC, 0x9999} ,
   {0xA5A5, 0x9696, 0xAA55} ,
   {0x9966, 0xA55A, 0x9669} ,
} ;

/* 
   polynome used for hash calculation 
*/
unsigned int IncaIpHashGenPolyb[8] = { 1, 1, 0, 0, 0, 0, 0, 0 };


/*******************************************************************************
*
* IncaIpSwitchModulateMac - Modulate the address based on the given modulate code
*     The result is a modulated MAC address (based on the orignial address) which
*     can be used for further hash calculation.
*
* PARAMETERS:
*     modulateCode:  3x 16 bit stream used for modulation
*     inAddrPart:    16 bit part of the original MAC address. The address part can be 
*                    a combination of the original MAC address.
*     retAddr:       retrun parameter, give back the modulated MAC address
*
*/
void IncaIpSwitchModulateMac(u16 modulateCode[3], unsigned char inAddrPart[2], unsigned char *retAddr)
{
	int inputb[16];
	int codeb[48], not_codeb[48];
	int signalb[48];
	
	int i, j, n;

   /* init inputb table */
	for (i = 0; i < 8; i++)
	{	
		inputb[8+i]  = ((inAddrPart[0] >> i) & 0x0001);
      inputb[i]    = ((inAddrPart[1] >> i) & 0x0001);   
   }        /* for (i = 0; i < 8; i++) {...} */

   /* init codeb and not_codeb array */
   for (i = 0; i < 16; i++)
	{	
		codeb[32+i] = (modulateCode[0] >> i) & 0x0001 ;
      not_codeb[32+i] = (codeb[32+i] == 1) ? 0:1 ;

		codeb[16+i] = (modulateCode[1] >> i) & 0x0001 ;
      not_codeb[16+i] = (codeb[16+i] == 1) ? 0:1 ;

		codeb[i] =    (modulateCode[2] >> i) & 0x0001 ;
      not_codeb[i] = (codeb[i] == 1) ? 0:1 ;
	}        /* for (i = 0; i < 16; i++) {...} */

   /* reset signalb array */
   memset(signalb, 0x00, sizeof(signalb) ) ;

	for (n = i = 0; i < 16; i++)
	{
		if (inputb[i] == 1)
		{
			for (j = 0; j < 3; j++, n++)
			{
				signalb[n] = codeb[n];
			}
		}
		else     /* if (inputb[i] == 1) {...} */
		{
			for (j = 0; j < 3; j++, n++)
			{
				signalb[n] = not_codeb[n];
			}
		}        /* if (inputb[i] == 1) {...} */
	}        /* for (n = i = 0; i < 16; i++) {...} */

   /* set return address */
   memset(retAddr, 0x00, sizeof(retAddr) ) ;

	for (i = 8; i >=0; i--)
	{
      retAddr[5] = retAddr[5] << 1 ;
      retAddr[4] = retAddr[4] << 1 ;
      retAddr[3] = retAddr[3] << 1 ;
      retAddr[2] = retAddr[2] << 1 ;
      retAddr[1] = retAddr[1] << 1 ;
      retAddr[0] = retAddr[0] << 1 ;

		if (signalb[i] == 1)
			retAddr[5] |= 0x01;
      if (signalb[i+8] == 1)
			retAddr[4] |= 0x01;

      if (signalb[i+16] == 1)
			retAddr[3] |= 0x01;
      if (signalb[i+24] == 1)
			retAddr[2] |= 0x01;

      if (signalb[i+32] == 1)
			retAddr[1] |= 0x01;
      if (signalb[i+40] == 1)
			retAddr[0] |= 0x01;
	}        /* for (i = 8; i >=0; i--) {...} */
}        /* LOCAL void IncaIpSwitchModulateMac(...) {...} */


/*******************************************************************************
*
* IncaIpSwitchCalcMacHash - Calculate a hash value form the given MAC address
*
* PARAMETERS:
*     inMacAddr:     MAC address used for hash calculation
*     len:           currently fixed 8
*
*  RETURN:
*     calculated hash value
*
*  REMARK:
*     This value is used as MAC table index for locate a MAC address within the
*     INCA-IP MAC table (switch module)
*/
unsigned char IncaIpSwitchCalcMacHash(unsigned char inMacAddr[6], u32 len)
{
	int   i, j ;
	int   inputb[48] ;
	int   crc_reg[8], crc[8] ;
	unsigned char crc_reg_final ;

   /* init inputb table */
	for (i = 0; i < 8; i++)
	{	
		inputb[40+i] = ((inMacAddr[0] >> i) & 0x0001);
		inputb[32+i] = ((inMacAddr[1] >> i) & 0x0001);
		inputb[24+i] = ((inMacAddr[2] >> i) & 0x0001);
		inputb[16+i] = ((inMacAddr[3] >> i) & 0x0001);
		inputb[8+i]  = ((inMacAddr[4] >> i) & 0x0001);
      inputb[i]    = ((inMacAddr[5] >> i) & 0x0001);
	}        /* for (i = 0; i < 8; i++) {...} */

   /* init CRC table */
	for(i = 0; i < 8; i++)
	{
		crc_reg[i] = 0;
		crc[i] = 0;
	}

 	for (i = 47; i >= 0; i--)
	{
		if ((crc_reg[len-1] ^ inputb[i]) == 1)
		{			
			crc_reg[len-1] = 0;
			for (j = 0; j < 8; j++)
			{
				crc[7-j] = crc_reg[7-j] ^ IncaIpHashGenPolyb[7-j];
			}
			
			for(j = 0; j < 7; j++)
			{
				crc[7-j] = crc[7-j-1];
			}
			crc[0] = 1;
		}
		else     /* if ((crc_reg[len-1] ^ inputb[i]) == 1) {...} else ... */
		{
			for(j = 0; j < 7; j++)
			{
				crc[7-j] = crc[7-j-1];
			}
			crc[0] = 0;
		}        /* if ((crc_reg[len-1] ^ inputb[i]) == 1) {...} else {...} */
		
		for (j = 0; j < 8; j++)
			crc_reg[j] = crc[j];
	}        /* for (i = 47; i >= 0; i--) {...} */

 	crc_reg_final = 0x00;
	for (i = 7; i >=0; i--)
	{
		crc_reg_final = crc_reg_final << 1;
		
		if (crc_reg[i] == 1)
			crc_reg_final = crc_reg_final | 0x01;
			
	}        /* for (i = 7; i >=0; i--) {...} */

   return (crc_reg_final) ;
}




/*******************************************************************************
*
* sw21553_IncaIpSwitchCalcHash - choose a modulation code and the MAC address part
*     used for calculation of the modulated MAC address
*
* Parameters:
*     modulateNum:   range 0..2 - chose the used modulation code
*     inMacAddr:     original MAC address From this address a modulation depended part 
*                    will be used for calculation
*
*  Return:
*     hash value: returns the calculated hash value
*     -1:         ERROR, invalid modulation number
*
*  Remark:
*     The MAC address is expected a simple byte stream (no endian)
*     example: 00:50:04:cb:18:8c 
*              --> unsigned char inMacAddr[6] = {0x00, 0x50, 0x04, 0xcb, 0x18, 0x8c}
*/
unsigned char sw21553_IncaIpSwitchCalcHash(unsigned int modulateNum, unsigned char inMacAddr[6])
{
   unsigned char  modulateMacAddrPart[2]  ; 
   unsigned char  modulateMacAddr[6] ;

   switch(modulateNum)
   {
      case 0:
         /* 2. HASH */
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[0] ^ inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[1] ^ inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_2, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 1:
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[2] ^ inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[3] ^ inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_1, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 2:
         /* calculate next modulated hash */
         modulateMacAddrPart[0] = inMacAddr[4] ;
         modulateMacAddrPart[1] = inMacAddr[5] ;
         memset(modulateMacAddr, 0x00, sizeof(modulateMacAddr) ) ;
         IncaIpSwitchModulateMac( IncaIpMacModulateCodes.Code_0, 
                                  &modulateMacAddrPart[0], 
                                  modulateMacAddr) ;
         return( IncaIpSwitchCalcMacHash(modulateMacAddr, 8) ) ;
         break ;
      case 3:
         /* calculate hash based on original MAC address */
         return( IncaIpSwitchCalcMacHash(inMacAddr, 8) ) ;
         break ;
      default:
         return(-1) ;
         break ;
   }        /* switch(modulateNum) {...} */
}        /* UCHAR sw21553_IncaIpSwitchCalcHash(...) {...} */




/******************************************************************************
*  sw21553_SwMacSetTableEntryCntrl()
*     interprete and set the given properties for the control byte within
*     a MAC table entry.
*
*  Parameter:
*     port:       defines the port to which the address is assigned
*                 possible values are:
*                 0: PC port
*                 1: LAN port
*                 2: CPU port VOICE
*                 3: CPU port MANAGEMENT
*     classification: calssify the frames
*                 0: packets will be classified as normal packets
*                 1: packets will be classified as critical packets
*     aging:      defines the aging behaviour of the entry
*                 0: no aging --> static entry into the MAC table
*                 1: normal aging --> entry will normal timed out
*     pCntrl:     ptr to the MAC entry control byte
*
*  Returns:
*     0 (OK):    given values are valid and control byte was set
*    -1 (ERROR): error while interpretation of the control byte
*                - no ptr for return of the value given
*                - port assignment not valid
*
*  REMARK:
*
*/
int sw21553_SwMacSetTableEntryCntrl(int port, int classification, int aging, unsigned char *pCntrl)
{
   unsigned char tempCntrl = 0;

   if (pCntrl == NULL)
      return(-1) ;
   /* check and set assigned port */
   switch(port)
   {
      case INCA_IP_SWITCH_PORT_SELECT_PC:
         tempCntrl = (unsigned char)INCA_IP_MAC_TABLE_ENTRY_PORT_PC_MASK ;
         break ;

      case INCA_IP_SWITCH_PORT_SELECT_LAN:
         tempCntrl = (unsigned char)INCA_IP_MAC_TABLE_ENTRY_PORT_LAN_MASK ;
         break ;

      case INCA_IP_SWITCH_PORT_SELECT_CPU_V:
         tempCntrl = (unsigned char)INCA_IP_MAC_TABLE_ENTRY_PORT_CPU_V_MASK ;
         break ;

      case INCA_IP_SWITCH_PORT_SELECT_CPU_M:
         tempCntrl = (unsigned char)INCA_IP_MAC_TABLE_ENTRY_PORT_CPU_M_MASK ;
         break ;

      default:
         return(-1) ;
   }        /* switch(port) {...} */

   tempCntrl |= (unsigned char)( (classification) ? (INCA_IP_MAC_TABLE_ENTRY_CRITICAL_MASK):(INCA_IP_MAC_TABLE_ENTRY_NO_CRITICAL_MASK) ) ;
   tempCntrl |= (unsigned char)( (aging) ? (INCA_IP_MAC_TABLE_ENTRY_MA_ST_NORMAL_AGING_MASK):(INCA_IP_MAC_TABLE_ENTRY_MA_ST_STATIC_MASK) ) ;

   /* set return value */
   *pCntrl = tempCntrl ;

   return(0) ;
}



/******************************************************************************
*  sw21553_SwMacReadTableEntry()
*     read the MAC entry from the MAC table
*
*  Parameter:
*     index: table index of the requested entry
*     pOutMacTableEntry: ptr to the return structure
*
*  Returns:
*     0: no entry found at this index
*     1: one entry found and returned
*    -1: ERROR ptr or index not valid 
*
*  REMARK:
*     The control parameter indicates the validation of the entry
*     concerning aging and port assignment
*/
int sw21553_SwMacReadTableEntry(int index, sw21553_MacTableEntry *pOutMacTableEntry )
{
   u32 readCmd ;
   sw21553_CdaRegister  CADReg1, CADReg2 ;

   if ( 
        (pOutMacTableEntry == NULL) ||
        (index < 0) || 
        (index >= INCA_IP_MAX_MAC_TABLE_ENTRIES) 
      )
   {
      return(-1) ;
   }

   readCmd = INCA_IP_Switch_CPU_ACTL_RD_MA_TABLE | index ;
   /* MAC table: location 0 */
   SW_WRITE_REG( INCA_IP_Switch_CPU_ACTL, readCmd ) ;
   while(1)
   {
      SW_READ_REG(INCA_IP_Switch_CPU_ACTL, readCmd) ;
      if ( (readCmd & 0x80000000) == 0 )
         break ;
   }

   SW_READ_REG(INCA_IP_Switch_DATA1, CADReg1.reg32) ;
   SW_READ_REG(INCA_IP_Switch_DATA2, CADReg2.reg32) ;

   if ( (CADReg1.reg32 != 0x00FFFFFF) || (CADReg2.reg32 != 0xFFFFFFFF) )
   {
      /* entry found --> transfere data */
      pOutMacTableEntry->Cntrl = CADReg1.reg8[1] ;
      pOutMacTableEntry->MacAddr[0] = CADReg1.reg8[2] ;
      pOutMacTableEntry->MacAddr[1] = CADReg1.reg8[3] ;
      pOutMacTableEntry->MacAddr[2] = CADReg2.reg8[0] ;
      pOutMacTableEntry->MacAddr[3] = CADReg2.reg8[1] ;
      pOutMacTableEntry->MacAddr[4] = CADReg2.reg8[2] ;
      pOutMacTableEntry->MacAddr[5] = CADReg2.reg8[3] ;
      return(1) ;
   }

   return(0) ;
}




/******************************************************************************
*  sw21553_SwMacWriteTableEntry()
*     read the MAC entry from the MAC table
*
*  Parameter:
*     index: table index where the MAC entry will be written
*     inMacTableEntry: MAC table entry
*
*  Returns:
*     0 (OK):    write of the new entry was successful
*    -1 (ERROR): write of the new entry was not successful
*
*  REMARK:
*     Index must be in range of valid indexes
*     The given MAC table entry must contain valid data
*/
int sw21553_SwMacWriteTableEntry(int index, sw21553_MacTableEntry inMacTableEntry )
{
   u32 writeCmd, count ;
   sw21553_CdaRegister  CADReg1, CADReg2 ;

   CADReg1.reg32 = 0 ;
   CADReg2.reg32 = 0 ;
   CADReg1.reg8[1] = inMacTableEntry.Cntrl ;
   CADReg1.reg8[2] = inMacTableEntry.MacAddr[0] ;
   CADReg1.reg8[3] = inMacTableEntry.MacAddr[1] ;

   CADReg2.reg8[0] = inMacTableEntry.MacAddr[2] ;
   CADReg2.reg8[1] = inMacTableEntry.MacAddr[3] ;
   CADReg2.reg8[2] = inMacTableEntry.MacAddr[4] ;
   CADReg2.reg8[3] = inMacTableEntry.MacAddr[5] ;

   /* write new MAC entry to the CPU access register */
   SW_WRITE_REG(INCA_IP_Switch_DATA1, CADReg1.reg32) ;
   SW_WRITE_REG(INCA_IP_Switch_DATA2, CADReg2.reg32) ;
/*
   printf("MAC entry: 0x%08X / 0x%08X written\n", 
          CADReg1.reg32, CADReg2.reg32);
*/   
   /* set write commend for the MAC table */
   writeCmd = INCA_IP_Switch_CPU_ACTL_WR_MA_TABLE | (u32)index ;
   SW_WRITE_REG( INCA_IP_Switch_CPU_ACTL, writeCmd ) ;

   /* check write */
   count = 0 ;
   while(1)
   {
      SW_READ_REG(INCA_IP_Switch_CPU_ACTL, writeCmd) ;
      if ( (writeCmd & 0x80000000) == 0 || count > 10)
         break ;

      count++ ;
   }        /* while(1) {...} */

   if ( (writeCmd & 0x80000000) == 0)
      return(0) ;
   else
      return(-1) ;
}



/******************************************************************************
*  sw21553_AddMacEntry()
*     Add an given MAC table entry to the MAC table.
*     The MAC table index will be calculated via the hash algorithm
*
*  Parameter:
*     inMacTableEntry: MAC address and appropriate control structures
*     overWrite:       add this entry always (overwrite existing)
*
*  Returns:
*     0 (OK):    add of the new entry was successful
*    -1 (ERROR): add of the new entry was not successful, 
*                calculated entries not free
*
*  REMARK:
*     The given MAC table entry must contain valid and already complete data
*/
int sw21553_SwMacAddEntry(sw21553_MacTableEntry inMacTableEntry, int overWrite)
{
   unsigned char  macTableHash ;
   unsigned int   modulateNum = 0 ;
   sw21553_MacTableEntry tempMacTableEntry ;

/*   printf("sw21553_SwMacAddEntry()\n");  */   
   
   for (modulateNum = 0 ; modulateNum <= 3 ; modulateNum++)
   {
      /* 
         calculate the MAC table index via hash algorithm 
         REMEBER: init of the polynome has to be done before 
                  --> sw21553_IncaIpInitHashGenPoly() ;
      */
      macTableHash = sw21553_IncaIpSwitchCalcHash(modulateNum, inMacTableEntry.MacAddr) ;

      if (overWrite == 0)
      {
         /* check if this entry is free */
         if ( sw21553_SwMacReadTableEntry(macTableHash, &tempMacTableEntry) == 1 )
         {
            /* table entry bussy --> check if this entry is still valid */
            if ( (tempMacTableEntry.Cntrl & (unsigned char)INCA_IP_MAC_TABLE_ENTRY_MA_ST_AGED_OUT_MASK) == 
                 INCA_IP_MAC_TABLE_ENTRY_MA_ST_AGED_OUT_MASK  )
            {
               /* entry is valid --> write entry */
               return (sw21553_SwMacWriteTableEntry((int)macTableHash, inMacTableEntry) ) ;
            }
         }
         else
         {
            /* table entry free --> write entry */
            return (sw21553_SwMacWriteTableEntry((int)macTableHash, inMacTableEntry) ) ;
         }
      }
      else
      {
         /* write always */
         return (sw21553_SwMacWriteTableEntry((int)macTableHash, inMacTableEntry) ) ;
      }
   }        /* for (modulateNum = 0 ; modulateNum < 3 ; modulateNum++) {...} */

   /* no free entry found */
   return(-1) ;
}





int inca_switch_initialize(bd_t * bis)
{
	struct eth_device *dev;

#if 0
	printf("Entered inca_switch_initialize()\n");
#endif

	if (!(dev = (struct eth_device *) malloc (sizeof *dev))) {
		printf("Failed to allocate memory\n");
		return 0;
	}
	memset(dev, 0, sizeof(*dev));

	inca_dma_init();
	inca_init_switch_chip();
#if defined(CONFIG_INCA_IP_SWITCH_AMDIX)
	inca_amdix();
#endif
	sprintf(dev->name, "INCA-IP Switch");

	inca_check_connections();

	dev->init = inca_switch_init;
	dev->halt = inca_switch_halt;
	dev->send = inca_switch_send;
	dev->recv = inca_switch_recv;

	eth_register(dev);

#if 0
	printf("Leaving inca_switch_initialize()\n");
#endif

	return 1;
}


static int inca_switch_init(struct eth_device *dev, bd_t * bis)
{
	int i;
	u32 v;
	sw21553_MacTableEntry tempMacTableEntry;    

#if 0
	printf("Entering inca_switch_init()\n");
#endif

	memcpy(tempMacTableEntry.MacAddr, dev->enetaddr, 6);
    	sw21553_SwMacSetTableEntryCntrl( INCA_IP_SWITCH_PORT_SELECT_CPU_V,   /* CPU voice port */
                                     1,                                  /* critical */
                                     0,                                  /* static entry */ 
                                     &tempMacTableEntry.Cntrl);
   	/* add own MAC address, overwrite existing */
   	sw21553_SwMacAddEntry(tempMacTableEntry, 1) ;    
 	
	/* Initialize the descriptor rings.
	 */
	for (i = 0; i < NUM_RX_DESC; i++) {
		inca_rx_descriptor_t * rx_desc = KSEG1ADDR(&rx_ring[i]);
		memset(rx_desc, 0, sizeof(rx_ring[i]));

		/* Set maximum size of receive buffer.
		 */
		rx_desc->params.field.NFB = PKTSIZE_ALIGN;

		/* Set the offset of the receive buffer. Zero means
		 * that the offset mechanism is not used.
		 */
		rx_desc->params.field.offset = 0;

		/* Check if it is the last descriptor.
		 */
		if (i == (NUM_RX_DESC - 1)) {
			/* Let the last descriptor point to the first
			 * one.
			 */
			rx_desc->nextRxDescPtr = KSEG1ADDR((u32)rx_ring);
		} else {
			/* Set the address of the next descriptor.
			 */
			rx_desc->nextRxDescPtr = (u32)KSEG1ADDR(&rx_ring[i+1]);
		}

		rx_desc->RxDataPtr = (u32)KSEG1ADDR(NetRxPackets[i]);
	}

#if 0
	printf("rx_ring = 0x%08X 0x%08X\n", (u32)rx_ring, (u32)&rx_ring[0]);
	printf("tx_ring = 0x%08X 0x%08X\n", (u32)tx_ring, (u32)&tx_ring[0]);
#endif

	for (i = 0; i < NUM_TX_DESC; i++) {
		inca_tx_descriptor_t * tx_desc = KSEG1ADDR(&tx_ring[i]);

		memset(tx_desc, 0, sizeof(tx_ring[i]));

		tx_desc->params.word	   = 0;
		tx_desc->params.field.HOLD = 1;
		tx_desc->C		   = 1;

			/* Check if it is the last descriptor.
			 */
		if (i == (NUM_TX_DESC - 1)) {
				/* Let the last descriptor point to the
				 * first one.
				 */
			tx_desc->nextTxDescPtr = KSEG1ADDR((u32)tx_ring);
		} else {
				/* Set the address of the next descriptor.
				 */
			tx_desc->nextTxDescPtr = (u32)KSEG1ADDR(&tx_ring[i+1]);
		}
	}

	/* Initialize RxDMA.
	 */
	DMA_READ_REG(INCA_IP_DMA_DMA_RXISR, v);
#if 0
	printf("RX status = 0x%08X\n", v);
#endif

	/* Writing to the FRDA of CHANNEL.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXFRDA0, (u32)rx_ring);

	/* Writing to the COMMAND REG.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_INIT);

	/* Initialize TxDMA.
	 */
	DMA_READ_REG(INCA_IP_DMA_DMA_TXISR, v);
#if 0
	printf("TX status = 0x%08X\n", v);
#endif

	/* Writing to the FRDA of CHANNEL.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXFRDA0, (u32)tx_ring);

	tx_new = rx_new = 0;

	tx_hold = NUM_TX_DESC - 1;
	rx_hold = NUM_RX_DESC - 1;

#if 0
	rx_ring[rx_hold].params.field.HOLD = 1;
#endif
	/* enable spanning tree forwarding, enable the CPU port */
	/* ST_PT:
	 *	CPS (CPU port status)	0x3 (forwarding)
	 *	LPS (LAN port status)	0x3 (forwarding)
	 *	PPS (PC port status)	0x3 (forwarding)
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0x3f);

#if 0
	printf("Leaving inca_switch_init()\n");
#endif

	return 0;
}


static int inca_switch_send(struct eth_device *dev, volatile void *packet, int length)
{
	int		       i;
	int		       res	= -1;
	u32		       command;
	u32		       regValue;
	inca_tx_descriptor_t * tx_desc	= KSEG1ADDR(&tx_ring[tx_new]);

#if 0
	printf("Entered inca_switch_send()\n");
#endif

	if (length <= 0) {
		printf ("%s: bad packet size: %d\n", dev->name, length);
		goto Done;
	}

	for(i = 0; tx_desc->C == 0; i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: tx error buffer not ready\n", dev->name);
			goto Done;
		}
	}

	if (tx_old_hold >= 0) {
		KSEG1ADDR(&tx_ring[tx_old_hold])->params.field.HOLD = 1;
	}
	tx_old_hold = tx_hold;

	tx_desc->params.word =
			(INCA_DMA_TX_SOP | INCA_DMA_TX_EOP | INCA_DMA_TX_HOLD);

	tx_desc->C = 0;
	tx_desc->TxDataPtr = (u32)packet;
	tx_desc->params.field.NBA = length;

	KSEG1ADDR(&tx_ring[tx_hold])->params.field.HOLD = 0;

	tx_hold = tx_new;
	tx_new	= (tx_new + 1) % NUM_TX_DESC;


	if (! initialized) {
		command = INCA_IP_DMA_DMA_TXCCR0_INIT;
		initialized = 1;
	} else {
		command = INCA_IP_DMA_DMA_TXCCR0_HR;
	}

	DMA_READ_REG(INCA_IP_DMA_DMA_TXCCR0, regValue);
	regValue |= command;
#if 0
	printf("regValue = 0x%x\n", regValue);
#endif
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXCCR0, regValue);

#if 0
	for(i = 0; KSEG1ADDR(&tx_ring[tx_hold])->C == 0; i++) {
		if (i >= TOUT_LOOP) {
			printf("%s: tx buffer not ready\n", dev->name);
			goto Done;
		}
	}
#endif
	res = length;
Done:
#if 0
	printf("Leaving inca_switch_send()\n");
#endif
	return res;
}


static int inca_switch_recv(struct eth_device *dev)
{
	int		       length  = 0;
	inca_rx_descriptor_t * rx_desc;

#if 0
	printf("Entered inca_switch_recv()\n");
#endif

	for (;;) {
		rx_desc = KSEG1ADDR(&rx_ring[rx_new]);
		/*  de6wmigr:  
            	To be checked why this delay prevents the ping from hanging */
        	udelay(10); 

		if (rx_desc->status.field.C == 0) {
			break;
		}

#if 0
		rx_ring[rx_new].params.field.HOLD = 1;
#endif

		if (! rx_desc->status.field.Eop) {
			printf("Partly received packet!!!\n");
			break;
		}

		length = rx_desc->status.field.NBT;
		rx_desc->status.word &=
			 ~(INCA_DMA_RX_EOP | INCA_DMA_RX_SOP | INCA_DMA_RX_C);
#if 0
{
  int i;
  for (i=0;i<length - 4;i++) {
    if (i % 16 == 0) printf("\n%04x: ", i);
    printf("%02X ", NetRxPackets[rx_new][i]);
  }
  printf("\n");
}
#endif

		if (length) {
#if 0
			printf("Received %d bytes\n", length);
#endif
			NetReceive((void*)KSEG1ADDR(NetRxPackets[rx_new]), length - 4);
		} else {
#if 1
			printf("Zero length!!!\n");
#endif
		}


		KSEG1ADDR(&rx_ring[rx_hold])->params.field.HOLD = 0;

		rx_hold = rx_new;

		rx_new = (rx_new + 1) % NUM_RX_DESC;
	}

#if 0
	printf("Leaving inca_switch_recv()\n");
#endif

	return length;
}


static void inca_switch_halt(struct eth_device *dev)
{
#if 0
	printf("Entered inca_switch_halt()\n");
#endif

#if 1
	initialized = 0;
#endif
#if 1
	/* de6wmigr:  For any unknown reason this delay prevents the ping
       command over the PC port from crashing */
    	udelay(10);

	/* Disable forwarding to the CPU port.
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf);

	/* Close RxDMA channel.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);

	/* Close TxDMA channel.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXCCR0, INCA_IP_DMA_DMA_TXCCR0_OFF);


#endif
#if 0
	printf("Leaving inca_switch_halt()\n");
#endif
}


static void inca_init_switch_chip(void)
{
	u32 regValue;

	/* To workaround a problem with collision counter
	 * (see Errata sheet).
	 */
	SW_WRITE_REG(INCA_IP_Switch_PC_TX_CTL, 0x00000001);
	SW_WRITE_REG(INCA_IP_Switch_LAN_TX_CTL, 0x00000001);

#if 1
	/* init MDIO configuration:
	 *	MDS (Poll speed):	0x01 (4ms)
	 *	PHY_LAN_ADDR:		0x06
	 *	PHY_PC_ADDR:		0x05
	 *	UEP (Use External PHY): 0x00 (Internal PHY is used)
	 *	PS (Port Select):	0x00 (PT/UMM for LAN)
	 *	PT (PHY Test):		0x00 (no test mode)
	 *	UMM (Use MDIO Mode):	0x00 (state machine is disabled)
	 */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_CFG, 0x4c50);

	/* init PHY:
	 *	SL (Auto Neg. Speed for LAN)
	 *	SP (Auto Neg. Speed for PC)
	 *	LL (Link Status for LAN)
	 *	LP (Link Status for PC)
	 *	DL (Duplex Status for LAN)
	 *	DP (Duplex Status for PC)
	 *	PL (Auto Neg. Pause Status for LAN)
	 *	PP (Auto Neg. Pause Status for PC)
	 */
	SW_WRITE_REG (INCA_IP_Switch_EPHY, 0xff);

	/* MDIO_ACC:
	 *	RA (Request/Ack)  0x01 (Request)
	 *	RW (Read/Write)	  0x01 (Write)
	 *	PHY_ADDR	  0x05 (PC)
	 *	REG_ADDR	  0x00 (PHY_BCR: basic control register)
	 *	PHY_DATA	  0x8000
	 *		      Reset		      - software reset
	 *		      LB (loop back)	      - normal
	 *		      SS (speed select)	      - 10 Mbit/s
	 *		      ANE (auto neg. enable)  - enable
	 *		      PD (power down)	      - normal
	 *		      ISO (isolate)	      - normal
	 *		      RAN (restart auto neg.) - normal
	 *		      DM (duplex mode)	      - half duplex
	 *		      CT (collision test)     - enable
	 */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC, 0xc0a09000);

	/* MDIO_ACC:
	 *	RA (Request/Ack)  0x01 (Request)
	 *	RW (Read/Write)	  0x01 (Write)
	 *	PHY_ADDR	  0x06 (LAN)
	 *	REG_ADDR	  0x00 (PHY_BCR: basic control register)
	 *	PHY_DATA	  0x8000
	 *		      Reset		      - software reset
	 *		      LB (loop back)	      - normal
	 *		      SS (speed select)	      - 10 Mbit/s
	 *		      ANE (auto neg. enable)  - enable
	 *		      PD (power down)	      - normal
	 *		      ISO (isolate)	      - normal
	 *		      RAN (restart auto neg.) - normal
	 *		      DM (duplex mode)	      - half duplex
	 *		      CT (collision test)     - enable
	 */
	SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC, 0xc0c09000);

#endif

	/* Make sure the CPU port is disabled for now. We
	 * don't want packets to get stacked for us until
	 * we enable DMA and are prepared to receive them.
	 */
	SW_WRITE_REG(INCA_IP_Switch_ST_PT,0xf);

	SW_READ_REG(INCA_IP_Switch_ARL_CTL, regValue);

	/* CRC GEN is enabled.
	 */
	regValue |= 0x00000200;
	SW_WRITE_REG(INCA_IP_Switch_ARL_CTL, regValue);

	/* ADD TAG is disabled.
	 */
	SW_READ_REG(INCA_IP_Switch_PMAC_HD_CTL, regValue);
	regValue &= ~0x00000002;
	SW_WRITE_REG(INCA_IP_Switch_PMAC_HD_CTL, regValue);
}


static void inca_dma_init(void)
{
	/* Switch off all DMA channels.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXCCR1, INCA_IP_DMA_DMA_RXCCR1_OFF);

	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXCCR0, INCA_IP_DMA_DMA_RXCCR0_OFF);
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXCCR1, INCA_IP_DMA_DMA_TXCCR1_OFF);
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXCCR2, INCA_IP_DMA_DMA_TXCCR2_OFF);

	/* Setup TX channel polling time.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXPOLL, INCA_DMA_TX_POLLING_TIME);

	/* Setup RX channel polling time.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXPOLL, INCA_DMA_RX_POLLING_TIME);

	/* ERRATA: write reset value into the DMA RX IMR register.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXIMR, 0xFFFFFFFF);

	/* Just in case: disable all transmit interrupts also.
	 */
	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXIMR, 0xFFFFFFFF);

	DMA_WRITE_REG(INCA_IP_DMA_DMA_TXISR, 0xFFFFFFFF);
	DMA_WRITE_REG(INCA_IP_DMA_DMA_RXISR, 0xFFFFFFFF);
}


static int inca_check_connections(void)
{
	u32 phyReg1 = 0;
	u32 phyReg4 = 0;
	u32 phyReg5 = 0;
	u32 phyReg6 = 0;
	u32 phyReg31 = 0;
	u32 regEphy = 0;
#if defined INCAIP_AUTO_MDIX   
	int mdi_flag;
#endif    
	int retries;
    int lan_port_link_up = 0;  /* link is not up by default*/
         
	/* Setup GPIO pins.
	 */
	*INCA_IP_AUTO_MDIX_LAN_PORTS_DIR    |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
	*INCA_IP_AUTO_MDIX_LAN_PORTS_ALTSEL |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);

	/* Set MDI mode.
	 */
	*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT &= ~(1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
#if defined INCAIP_AUTO_MDIX       
	mdi_flag = 1;
#endif
	/* Wait for link.
	 */
	retries = WAIT_LINK_RETRIES;
	while (--retries  && !lan_port_link_up) 
    {
        udelay(LINK_RETRY_DELAY * 1000);  /* wait for negotiation completion */   
		SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x6 << 21) |	/* LAN		*/
				(1   << 16));	/* PHY_BSR	*/
		do {
			SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg1);
		} while (phyReg1 & (1 << 31));

        /* 
         * Set connection parameters if link is up
         */
		if (phyReg1 & (1 << 2)) 
        {
			/* Link is up */
//            printf("\nLink on LAN port is up !\n");
            lan_port_link_up = 1;
            
            /* Auto-negotiation / Parallel detection complete
		     */
		    if (phyReg1 & (1 << 5)) 
            {
//			    printf("ANEG/PARDET completed.\n");
                SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
                             (0x1 << 31) |	/* RA		*/
                             (0x0 << 30) |	/* Read		*/
                             (0x6 << 21) |	/* LAN		*/
                             (31  << 16));	/* PHY_SCSR	*/
			    do 
                {
                    SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg31);
                } while (phyReg31 & (1 << 31));

			    switch ((phyReg31 >> 2) & 0x7) 
                {
 			        case INCA_SWITCH_PHY_SPEED_10H:
//                        printf("10Base-T Half-duplex set for LAN port.\n");
                        /* 10Base-T Half-duplex */
 				        regEphy = 0;
 				        break;
 			        case INCA_SWITCH_PHY_SPEED_10F:
//                        printf("10Base-T Full-duplex set for LAN port.\n");
                        /* 10Base-T Full-duplex */
 				        regEphy = INCA_IP_Switch_EPHY_DL;
 				        break;
 			        case INCA_SWITCH_PHY_SPEED_100H:
//                        printf("100Base-TX Half-duplex set for LAN port.\n");
 				        /* 100Base-TX Half-duplex */
 				        regEphy = INCA_IP_Switch_EPHY_SL;
 				        break;
 			        case INCA_SWITCH_PHY_SPEED_100F:
//                        printf("100Base-TX Full-duplex set for LAN port.\n");
 				        /* 100Base-TX Full-duplex */
 				        regEphy = INCA_IP_Switch_EPHY_SL | INCA_IP_Switch_EPHY_DL;
 				        break;
                    default:
                        printf("Invalid speed setting for LAN port negotiated.\n");
                        break;
 			    }

                /* In case of Auto-negotiation,
                 * update the negotiated PAUSE support status
                 */
			    if (phyReg1 & (1 << 3)) 
                {
//				    printf("We are Auto Negotiation able.\n");
                    SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
					    (0x1 << 31) |	/* RA		*/
					    (0x0 << 30) |	/* Read		*/
					    (0x6 << 21) |	/* LAN		*/
					    (6   << 16));	/* PHY_ANER	*/
				    do 
                    {
                        SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg6);
                    } while (phyReg6 & (1 << 31));

                    /* We are Autoneg-able.
                     * Is Link partner also able to autoneg?
                     */
                    if (phyReg6 & (1 << 0)) 
                    {
//                        printf("Link Partner is Auto Negotiation able.\n");
                        SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
                                     (0x1 << 31) |	/* RA		*/
                                     (0x0 << 30) |	/* Read		*/
                                     (0x6 << 21) |	/* LAN		*/
                                     (4   << 16));	/* PHY_ANAR	*/
                        do 
                        {
                            SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg4);
                        } while (phyReg4 & (1 << 31));

                        /* We advertise PAUSE capab.
                         * Does link partner also advertise it?
                         */
                        if (phyReg4 & (1 << 10)) 
                        {
//                            printf("We are PAUSE capable.\n");
                            SW_WRITE_REG( INCA_IP_Switch_MDIO_ACC,
                                          (0x1 << 31) |	/* RA		*/
                                          (0x0 << 30) |	/* Read		*/
                                          (0x6 << 21) |	/* LAN		*/
                                          (5   << 16));	/* PHY_ANLPAR	*/
                            do 
                            {
                                SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg5);
                            } while (phyReg5 & (1 << 31));

                            /* Link partner is PAUSE capab.
                             */
                           if (phyReg5 & (1 << 10)) 
                           {
//                               printf("Link Partner is PAUSE capable.\n");
                               regEphy |= INCA_IP_Switch_EPHY_PL;
                           }
//                           else
//                               printf("Link Partner is not PAUSE capable.\n");
                        }
//                        else
//                            printf("We are not PAUSE capable.\n");
                    }
//                    else
//                        printf("Link Partner not Auto Negotiation able.\n");
                }
//                else
//                    printf("We are not Auto Negotiation able.\n");
            }
			/* Link is up */
			regEphy |= INCA_IP_Switch_EPHY_LL;
		}  /* end of link is up*/

     
        else  /* link is down -> toggle relais and await connection setup */
        {
#if defined INCAIP_AUTO_MDIX     
            if (mdi_flag) 
            {
			    /* Set MDIX mode */
                *INCA_IP_AUTO_MDIX_LAN_PORTS_OUT |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
                mdi_flag = 0;
            }
            else 
            {
                /* Set MDI mode */
                *INCA_IP_AUTO_MDIX_LAN_PORTS_OUT &= ~(1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
                mdi_flag = 1;
            }
#endif
        }
	}  /*   end of while( --retries) */

    
    /* check PC port connection */
    SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x5 << 21) |	/* PC		*/
				(1   << 16));	/* PHY_BSR	*/
	do {
		SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg1);
	} while (phyReg1 & (1 << 31));
   
	if (phyReg1 & (1 << 2)) 
    {
		/* Link is up */
//        printf("\nLink on PC port is up !\n");

        /* Auto-negotiation / Parallel detection complete
	     */
	    if (phyReg1 & (1 << 5)) 
        {
//		    printf("ANEG/PARDET completed.\n");
            SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
                         (0x1 << 31) |	/* RA		*/
                         (0x0 << 30) |	/* Read		*/
                         (0x5 << 21) |	/* PC		*/
                         (31  << 16));	/* PHY_SCSR	*/
		    do 
            {
                SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg31);
            } while (phyReg31 & (1 << 31));

		    switch ((phyReg31 >> 2) & 0x7) 
            {
		        case INCA_SWITCH_PHY_SPEED_10H:
//                    printf("10Base-T Half-duplex set for PC port.\n");
                    /* 10Base-T Half-duplex */
			        // regEphy = 0;
			        break;
		        case INCA_SWITCH_PHY_SPEED_10F:
//                    printf("10Base-T Full-duplex set for PC port.\n");
                    /* 10Base-T Full-duplex */
			        regEphy |= INCA_IP_Switch_EPHY_DP;
			        break;
		        case INCA_SWITCH_PHY_SPEED_100H:
//                    printf("100Base-TX Half-duplex set for PC port.\n");
			        /* 100Base-TX Half-duplex */
			        regEphy |= INCA_IP_Switch_EPHY_SP;
			        break;
		        case INCA_SWITCH_PHY_SPEED_100F:
//                    printf("100Base-TX Full-duplex set for PC port.\n");
			        /* 100Base-TX Full-duplex */
			        regEphy |= INCA_IP_Switch_EPHY_SP | INCA_IP_Switch_EPHY_DP;
			        break;
                default:
                    printf("Invalid speed setting for PC port negotiated.\n");
                    break;
		    }
            /* In case of Auto-negotiation,
             * update the negotiated PAUSE support status
             */
		    if (phyReg1 & (1 << 3)) 
            {
//			    printf("We are Auto Negotiation able.\n");
                SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				    (0x1 << 31) |	/* RA		*/
				    (0x0 << 30) |	/* Read		*/
				    (0x5 << 21) |	/* PC		*/
				    (6   << 16));	/* PHY_ANER	*/
			    do 
                {
                    SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg6);
                } while (phyReg6 & (1 << 31));

                /* We are Autoneg-able.
                 * Is Link partner also able to autoneg?
                 */
                if (phyReg6 & (1 << 0)) 
                {
//                    printf("Link Partner is Auto Negotiation able.\n");
                    SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
                                 (0x1 << 31) |	/* RA		*/
                                 (0x0 << 30) |	/* Read		*/
                                 (0x5 << 21) |	/* PC		*/
                                 (4   << 16));	/* PHY_ANAR	*/
                    do 
                    {
                        SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg4);
                    } while (phyReg4 & (1 << 31));

                    /* We advertise PAUSE capab.
                     * Does link partner also advertise it?
                     */
                    if (phyReg4 & (1 << 10)) 
                    {
                        SW_WRITE_REG( INCA_IP_Switch_MDIO_ACC,
                                      (0x1 << 31) |	/* RA		*/
                                      (0x0 << 30) |	/* Read		*/
                                      (0x5 << 21) |	/* PC		*/
                                      (5   << 16));	/* PHY_ANLPAR	*/
                        do 
                        {
                            SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg5);
                        } while (phyReg5 & (1 << 31));

                        /* Link partner is PAUSE capab.
                         */
                       if (phyReg5 & (1 << 10)) 
                       {
                           regEphy |= INCA_IP_Switch_EPHY_PP;
                       }
                   }
               }
//               else
//                   printf("Link Partner is not Auto Negotiation able.\n");            
            }
//            else
//                printf("We are not Auto Negotiation able.\n"); 
        }
           
        
		/* Link is up */
		regEphy |= INCA_IP_Switch_EPHY_LP;
	} 
//    else     
//        printf("\nNo Link on PC port !\n");
   
    /* update EPHY register with connection states read from phy registers */
    SW_WRITE_REG(INCA_IP_Switch_EPHY, regEphy);

	if (!lan_port_link_up) 
    {
//	    printf("\nNo Link on LAN port\n");
        return -1;		
	}
    else
    {
	    return 0;
    }

}
/* #if (CONFIG_COMMANDS & .. */



#if defined(CONFIG_INCA_IP_SWITCH_AMDIX)
static int inca_amdix(void)
{
	u32 phyReg1 = 0;
	u32 phyReg4 = 0;
	u32 phyReg5 = 0;
	u32 phyReg6 = 0;
	u32 phyReg31 = 0;
	u32 regEphy = 0;
	int mdi_flag;
	int retries;

	/* Setup GPIO pins.
	 */
	*INCA_IP_AUTO_MDIX_LAN_PORTS_DIR    |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
	*INCA_IP_AUTO_MDIX_LAN_PORTS_ALTSEL |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);

#if 0
	/* Wait for signal.
	 */
	retries = WAIT_SIGNAL_RETRIES;
	while (--retries) {
		SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x6 << 21) |	/* LAN		*/
				(17  << 16));	/* PHY_MCSR	*/
		do {
			SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg1);
		} while (phyReg1 & (1 << 31));

		if (phyReg1 & (1 << 1)) {
			/* Signal detected */
			break;
		}
	}

	if (!retries)
		goto Fail;
#endif

	/* Set MDI mode.
	 */
	*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT &= ~(1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
	mdi_flag = 1;

	/* Wait for link.
	 */
	retries = WAIT_LINK_RETRIES;
	while (--retries) {
		udelay(LINK_RETRY_DELAY * 1000);
		SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x6 << 21) |	/* LAN		*/
				(1   << 16));	/* PHY_BSR	*/
		do {
			SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg1);
		} while (phyReg1 & (1 << 31));

		if (phyReg1 & (1 << 2)) {
			/* Link is up */
			break;
		} else if (mdi_flag) {
			/* Set MDIX mode */
			*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT |= (1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
			mdi_flag = 0;
		} else {
			/* Set MDI mode */
			*INCA_IP_AUTO_MDIX_LAN_PORTS_OUT &= ~(1 << INCA_IP_AUTO_MDIX_LAN_GPIO_PIN_RXTX);
			mdi_flag = 1;
		}
	}

	if (!retries) {
		goto Fail;
	} else {
		SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x6 << 21) |	/* LAN		*/
				(1   << 16));	/* PHY_BSR	*/
		do {
			SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg1);
		} while (phyReg1 & (1 << 31));

		/* Auto-negotiation / Parallel detection complete
		 */
		if (phyReg1 & (1 << 5)) {
			SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
				(0x1 << 31) |	/* RA		*/
				(0x0 << 30) |	/* Read		*/
				(0x6 << 21) |	/* LAN		*/
				(31  << 16));	/* PHY_SCSR	*/
			do {
				SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg31);
			} while (phyReg31 & (1 << 31));

			switch ((phyReg31 >> 2) & 0x7) {
			case INCA_SWITCH_PHY_SPEED_10H:
				/* 10Base-T Half-duplex */
				regEphy = 0;
				break;
			case INCA_SWITCH_PHY_SPEED_10F:
				/* 10Base-T Full-duplex */
				regEphy = INCA_IP_Switch_EPHY_DL;
				break;
			case INCA_SWITCH_PHY_SPEED_100H:
				/* 100Base-TX Half-duplex */
				regEphy = INCA_IP_Switch_EPHY_SL;
				break;
			case INCA_SWITCH_PHY_SPEED_100F:
				/* 100Base-TX Full-duplex */
				regEphy = INCA_IP_Switch_EPHY_SL | INCA_IP_Switch_EPHY_DL;
				break;
			}

			/* In case of Auto-negotiation,
			 * update the negotiated PAUSE support status
			 */
			if (phyReg1 & (1 << 3)) {
				SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
					(0x1 << 31) |	/* RA		*/
					(0x0 << 30) |	/* Read		*/
					(0x6 << 21) |	/* LAN		*/
					(6   << 16));	/* PHY_ANER	*/
				do {
					SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg6);
				} while (phyReg6 & (1 << 31));

				/* We are Autoneg-able.
				 * Is Link partner also able to autoneg?
				 */
				if (phyReg6 & (1 << 0)) {
					SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
						(0x1 << 31) |	/* RA		*/
						(0x0 << 30) |	/* Read		*/
						(0x6 << 21) |	/* LAN		*/
						(4   << 16));	/* PHY_ANAR	*/
					do {
						SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg4);
					} while (phyReg4 & (1 << 31));

					/* We advertise PAUSE capab.
					 * Does link partner also advertise it?
					 */
					if (phyReg4 & (1 << 10)) {
						SW_WRITE_REG(INCA_IP_Switch_MDIO_ACC,
							(0x1 << 31) |	/* RA		*/
							(0x0 << 30) |	/* Read		*/
							(0x6 << 21) |	/* LAN		*/
							(5   << 16));	/* PHY_ANLPAR	*/
						do {
							SW_READ_REG(INCA_IP_Switch_MDIO_ACC, phyReg5);
						} while (phyReg5 & (1 << 31));

						/* Link partner is PAUSE capab.
						 */
						if (phyReg5 & (1 << 10)) {
							regEphy |= INCA_IP_Switch_EPHY_PL;
						}
					}
				}

			}

			/* Link is up */
			regEphy |= INCA_IP_Switch_EPHY_LL;

			SW_WRITE_REG(INCA_IP_Switch_EPHY, regEphy);
		}
	}

	return 0;

Fail:
	printf("No Link on LAN port\n");
	return -1;
}
#endif /* CONFIG_INCA_IP_SWITCH_AMDIX */

#endif
