#! /usr/bin/perl -w
#*****************************************************************************
#
#	Project:	Infineon CPU Subsystem
#	Description:	Convertting S-records to Verilog input format
#
##############################################################################
#
# mips_start_of_legal_notice
# 
# Copyright (c) 2001 MIPS Technologies, Inc. All rights reserved.
#
# Unpublished rights reserved under the Copyright Laws of the United States
# of America.
#
# This document contains information that is confidential and proprietary to
# MIPS Technologies, Inc. ("MIPS Technologies") and may be disclosed only to
# those employees of MIPS Technologies with a need to know, or as otherwise
# permitted in writing by MIPS Technologies. Any copying, modifying, use or
# disclosure of this information (in whole or in part) which is not
# expressly permitted in writing by MIPS Technologies is strictly
# prohibited. At a minimum, this information is protected under trade secret
# and unfair competition laws and the expression of the information
# contained herein is protected under federal copyright laws. Violations
# thereof may result in criminal penalties and fines.
# MIPS Technologies reserves the right to change the information contained
# in this document to improve function, design or otherwise.
# MIPS Technologies does not assume any liability arising out of the
# application or use of this information. Any license under patent rights or
# any other intellectual property rights owned by MIPS Technologies or third
# parties shall be conveyed by MIPS Technologies in a separate license
# agreement between MIPS Technologies and the licensee.
# The information contained in this document may only be disclosed to the
# U.S. Government, or to U.S. Government users, with prior written consent
# from MIPS Technologies.
# The information contained in this document constitutes one or more of the
# following: commercial computer software, commercial computer software
# documentation or other commercial items. If the user of this information,
# or any related documentation of any kind, including related technical data
# or manuals, is an agency, department, or other entity of the United States
# government ("Government"), the use, duplication, reproduction, release,
# modification, disclosure, or transfer of this information, or any related
# documentation of any kind, is restricted in accordance with Federal
# Acquisition Regulation 12.212 for civilian agencies and Defense Federal
# Acquisition Regulation Supplement 227.7202 for military agencies. The use
# of this information by the Government is further restricted in accordance
# with the terms of the license agreement(s) and/or applicable contract
# terms and conditions covering this information from MIPS Technologies.
# 
# mips_end_of_legal_notice
# 
#
#*****************************************************************************

use Getopt::Long;

$srec_endian = "EB";
$addr_bits   = 16;
$wordsize    = 64;
$miscinfo    = "";

$help = "
   Usage: $0 [-ES <endian>] [-A <addr bits>] [-W <word size> ] [-M <misc> ] <srecfile> <ofile>

   The script reads the <srecfile>.rec file which must be in SREC format,
   and writes two files out:
   <ofile>.asc     : File for readmem routine in Verilog.
   <ofile>_cfg.asc : Info on endianess and misc from -M option

   The following options are available:

   -ES <endian> : Set the endianess of the SREC file. Valid options:
                  -ES EL (little endian) or -ES EB (big endian)

   -A <addr bits> : Set number of address bits to include in ASC file.

   -W <word size> : Set the word size in output file.

   -M <misc>      : Additional data to write to the <file>_cfg.asc output file\n\n
";

GetOptions("ES:s"   => \$srec_endian,
     	   "A:s"    => \$addr_bits,
	   "W:s"    => \$wordsize,
	   "M:s"    => \$miscinfo);


if (!defined ($ARGV[0]) || !defined ($ARGV[1])) {
    print $help;
    exit -1;
}

$input_file      = $ARGV[0] . ".srec";
$output_file_asc = $ARGV[1] . ".asc";
$output_file_cfg = $ARGV[1] . "_cfg.asc";


open (IFILE, "<$input_file") ||
    die("Error: can't open file $input_file \n");

while ($line = <IFILE>)
{
    if ($line =~ /S3(.{2})(.{8})(.*)/)
    {
	$count = hex($1);
	$addr  = hex($2);
	$data  = $3;

	for ($i = 0; $i < $count-4-1; $i++)
	{
	    $data =~ /(.{2})/g;
	    $data_array{$addr} = $1;
	    $addr = $addr + 1;
	}
    }
}

close(IFILE);

#------------------------------------------------------------
# Create the ASC output
#------------------------------------------------------------
open (OFILE, ">$output_file_asc") ||
    die("Error: can't open file $output_file_asc \n");

$prev_word_addr = -256;
$bytes_per_word = $wordsize / 8;

# $addr_mask = 0xffffffff >> (32 - $addr_bits);
$addr_mask = (2 ** $addr_bits) - 1;

if    ($wordsize == 64) {$addr_shift = 3}
elsif ($wordsize == 32) {$addr_shift = 2}
elsif ($wordsize == 16) {$addr_shift = 1}
else                    {$addr_shift = 0}

for ($i=0; $i<$bytes_per_word; $i++)
{
    $output_word[$i] = "XX";
}

foreach $addr (sort sort_num keys(%data_array))
{
    $word_addr = ($addr >> $addr_shift) & $addr_mask;

    $byte_addr = $addr & (7 >> (3-$addr_shift));

    # We have entered a new word
    if ($word_addr != $prev_word_addr)
    {

	# Write out word, unless this is the very first
	# time we are here.
	if ($prev_word_addr != -256)
	{
	    foreach $w (@output_word) { print (OFILE $w);}
	    print (OFILE "\n");

	    # Reset word
	    for ($i=0; $i<$bytes_per_word; $i++)
	    {
		$output_word[$i] = "XX";
	    }

	}

	# If we have a hole in the address space.
	if ($word_addr != $prev_word_addr + 1)
	{
	    printf(OFILE "@%.8x\n",$word_addr);
	}

	$prev_word_addr = $word_addr;
    }

    if ($srec_endian eq "EB")
    {
	$output_word[$byte_addr] = $data_array{$addr};
    }
    else
    {
	$output_word[$bytes_per_word - $byte_addr - 1] = $data_array{$addr};
    }
}

# Write out the last word..
foreach $w (@output_word) { print (OFILE $w);}

close (OFILE);

#------------------------------------------------------------
# Create the CFG output file
#------------------------------------------------------------
open (OFILE, ">$output_file_cfg") ||
    die("Error: can't open file $output_file_cfg \n");

print (OFILE "// Endian (-ES): $srec_endian\n");
print (OFILE "// Misc   (-M) : $miscinfo\n");

if ($srec_endian eq "EB")
{
    print (OFILE "1$miscinfo\n");
}
else
{
    print (OFILE "0$miscinfo\n");
}

sub sort_num {
    if ($a > $b) { $retval = 1; }
    elsif ($a == $b) { $retval = 0; }
    else { $retval = -1; }
    $retval;
}

exit 0;












